from izigraph.importers import BaseImporter
from izigraph import Graph


class NgvImporter(BaseImporter):

    def import_graph(self, to_import):
        g = Graph(node_label_format="{0}", directed=False)
        nodes, links = self.__get_parsed_graph_parameters(to_import)
        for node in nodes:
            g.add_node(x=node[0], y=node[1])
        for link in links:
            g.link_nodes(link[0], link[1], weight_by_distance=True)
        return g

    def __get_parsed_graph_parameters(self, filename):
        parameters = self.__get_graph_parameters(filename)
        nodes_count = int(parameters[0])
        link_count = 0
        nodes = []
        links = []
        if len(parameters) > 1:
            link_count = int(parameters[1])
        first_node_index = 2
        last_node_index = first_node_index + nodes_count
        for node_index in range(first_node_index, last_node_index):
            nodes_raw_str = parameters[node_index]
            nodes_str = nodes_raw_str.replace(",", ".").split(";")
            node_position = (float(nodes_str[0]), float(nodes_str[1]))
            nodes.append(node_position)
        first_link_index = last_node_index
        last_link_index = first_link_index + link_count
        for link_index in range(first_link_index, last_link_index):
            links_raw_str = parameters[link_index]
            links_list = links_raw_str.split(";")
            links_tuple = (links_list[0], links_list[1])
            links.append(links_tuple)
        return nodes, links

    def __get_graph_parameters(self, filename):
        graph_str = self.__get_string_from_file(filename)
        parameters = graph_str.split("\n")
        return [p for p in parameters if not p == ""]

    def __get_string_from_file(self, filename):
        with open(filename, "r") as f:
            result = f.read()
        return result



try:
    from collections import OrderedDict
except ImportError:
    from ordereddict import OrderedDict
import izigraph


class Graph:
    def __init__(self, label='default_graph', node_label_format='node_{0}', directed=True):
        self.__label = label
        self.__nodes = {}
        self.__node_neighbors = {}
        self.__links_dict = {}
        self.__links = []
        self.__new_node_index = 0
        self.__node_label_format = node_label_format
        self.__is_directed = directed

    def label(self):
        """
        :return:The label of the Graph
        """
        return self.__label

    def set_label(self, label):
        """
        Changes the label of the graph
        :param label: string
        """
        self.__label = label

    def nodes(self):
        """
        Return the dictionary of Nodes. The key is the id of the node
        :return: dictionary of nodes
        """
        return OrderedDict(sorted(self.__nodes.items(), key=lambda t: t[0]))

    def links(self):
        return list(self.links_dict().values())

    def links_count(self):
        if self.__is_directed:
            return len(self.__links_dict)
        return len(self.__links_dict)/2

    def find_node_by_label(self, label):
        """
        :param label: The label string to search in the nodes_label dictionary
        :return: izigraph.node.Node or None
        """
        if not isinstance(label, str):
            raise ValueError('Label must be an instance of str')
        if label not in self.__nodes:
            return None
        else:
            return self.__nodes[label]

    def add_node(self, x=0.0, y=0.0):
        """
        :return:The added node
        """
        label = self.__node_label_format.format(self.__new_node_index)
        n = izigraph.Node(label, x=x, y=y)
        self.__nodes[n.label()] = n
        self.__new_node_index += 1
        return n

    def add_nodes(self, nodes):
        for i in range(self.__new_node_index, self.__new_node_index + nodes):
            self.add_node()

    def nodes_count(self):
        """
        :return:The size of the nodes dictionary
        """
        return len(self.__nodes)

    def link_nodes(self, source, destination, weight=1, weights=None, bidirectional=False, weight_by_distance=False):
        """
        :param weight_by_distance:
        :param source: izigraph.node.Node
        :param destination: izigraph.node.Node
        :param weight: int
        :param weights: list of int
        :param bidirectional: boolean
        :return:izigraph.link.Link or tuple of izigraph.link.Link
        """
        if isinstance(source, str):
            source = self.find_node_by_label(source)
        if isinstance(destination, str):
            destination = self.find_node_by_label(destination)
        self.validate_node(source)
        self.validate_node(destination)
        if source.label() not in self.__node_neighbors:
            self.__node_neighbors[source.label()] = []
        if not self.__is_directed and destination.label() not in self.__node_neighbors:
            self.__node_neighbors[destination.label()] = []
        if bidirectional:
            if weights is None:
                weights = [1, 1]
            if not isinstance(weights, list) and len(weights) != 2:
                raise ValueError('weights must be a list of two elements')
            if destination.label() not in self.__node_neighbors:
                self.__node_neighbors[destination.label()] = []
            s_d_link = source.link_to(destination, weight=weights[0])
            d_s_link = destination.link_to(source, weight=weights[1])
            self.__links_dict[(source.label(), destination.label())] = s_d_link
            self.__links_dict[(destination.label(), source.label())] = d_s_link
            self.__node_neighbors[source.label()].append(destination)
            self.__node_neighbors[destination.label()].append(source)
            return s_d_link, d_s_link
        if weight_by_distance:
            weight = self.get_distance_between_nodes(source, destination)
        s_d_link = source.link_to(destination, weight=weight)
        self.__links_dict[(source.label(), destination.label())] = s_d_link
        self.__node_neighbors[source.label()].append(destination)
        if not self.__is_directed:
            self.__links_dict[(destination.label(), source.label())] = destination.link_to(source, weight=weight)
            self.__node_neighbors[destination.label()].append(source)
        return s_d_link

    def node_belongs_to_graph(self, n):
        if isinstance(n, str):
            return n in self.__nodes
        elif isinstance(n, izigraph.Node):
            return n.label() in self.__nodes
        raise ValueError('n must be an instance of string or Node')

    def validate_node(self, n):
        if not isinstance(n, izigraph.Node):
            raise ValueError('node must be a izigraph.Node instance')
        if not self.node_belongs_to_graph(n):
            raise Exception('node doesn\'t belongs to graph {0}'.format(self.__label))

    def nodes_are_linked(self, source, destination):
        if isinstance(source, str):
            source = self.find_node_by_label(source)
        if isinstance(destination, str):
            destination = self.find_node_by_label(destination)
        return (source.label(), destination.label()) in self.__links_dict

    def get_link(self, source, destination):
        if isinstance(source, str):
            source = self.find_node_by_label(source)
        if isinstance(destination, str):
            destination = self.find_node_by_label(destination)
        if not self.nodes_are_linked(source, destination):
            return None
        return self.__links_dict[(source.label(), destination.label())]

    def is_directed(self):
        return self.__is_directed

    def links_dict(self):
        return OrderedDict(sorted(self.__links_dict.items(), key=lambda link: link[0]))

    def get_distance_between_nodes(self, source, destination):
        from math import sqrt, pow
        if isinstance(source, str):
            source = self.find_node_by_label(source)
        if isinstance(destination, str):
            destination = self.find_node_by_label(destination)
        x_distance = destination.position()[0] - source.position()[0]
        y_distance = destination.position()[1] - source.position()[1]
        distance_sum = pow(x_distance, 2) + pow(y_distance, 2)
        return sqrt(distance_sum)

    def unlink_nodes(self, source, destination):
        if isinstance(source, str):
            source = self.find_node_by_label(source)
        if isinstance(destination, str):
            destination = self.find_node_by_label(destination)
        if self.nodes_are_linked(source, destination):
            del(self.__links_dict[source.label(), destination.label()])
            source.unlink_from(destination)
            if not self.__is_directed:
                del (self.__links_dict[destination.label(), source.label()])
                destination.unlink_from(source)
            return True
        return False

    def node_neighbors(self):
        return OrderedDict(sorted(self.__node_neighbors.items(), key=lambda t: t[0]))

    def get_neighbors_from_node(self, node):
        if isinstance(node, str):
            return self.__node_neighbors[node]
        return self.__node_neighbors[node.label()]



import izigraph


class Link:
    def __init__(self, source, destination, weight=1):
        if not isinstance(weight, (int, float)):
            raise ValueError
        if not isinstance(source, izigraph.Node) or not isinstance(destination, izigraph.Node):
            raise ValueError
        self.__source = source
        self.__destination = destination
        self.__weight = weight

    def source(self):
        """
        Return the source node
        :return:izigraph.node.Node
        """
        return self.__source

    def destination(self):
        """
        Return the destination node
        :return:izigraph.node.Node
        """
        return self.__destination

    def weight(self):
        """
        Returns the weight of the Link
        :return:izigraph.node.Node
        """
        return self.__weight



try:
    from collections import OrderedDict
except ImportError:
    from ordereddict import OrderedDict
import izigraph


class Node:
    def __init__(self, label, x=0.0, y=0.0, z=0.0):
        self.__label = label
        self.__links = {}
        self.__x = x
        self.__y = y
        self.__z = z

    def label(self):
        """
        Returns the label of the node
        :return:str
        """
        return self.__label

    def position(self):
        return self.__x, self.__y

    def set_position(self, x=None, y=None, z=None):
        if x is not None:
            self.__x = x
        if y is not None:
            self.__y = y
        if z is not None:
            self.__z = z

    def links(self):
        """
        Returns the links of the node
        :return:list
        """
        return OrderedDict(self.__links)

    def nearest(self, exclude=None):
        links = self.links()
        nearest = (None, float('inf'))
        if exclude is None:
            exclude = []
        for link_label, link in links.items():
            current_weight = link.weight()
            if current_weight < nearest[1] and link_label not in exclude:
                nearest = (link.destination(), current_weight)
        return nearest

    def set_label(self, label):
        """
        Changes the label of the node
        :param label: str
        """
        self.__label = label

    def link_to(self, node, weight):
        """
        Generates a link with another node
        :param node: izigraph.node.Node
        :param weight: int or float
        :return:izigraph.link.Link
        """
        if self.is_linked_to(node):
            raise Exception("Nodes are already linked")
        l = izigraph.Link(self, node, weight)
        self.__links[str(node.label())] = l
        return l

    def is_linked_to(self, node):
        """
        Checks if the current node has a link wiht another node
        :param node: izigraph.node.Node
        :return:boolean
        """
        if str(node.label()) in self.__links:
            return True
        return False

    def get_link(self, node):
        """
        Returns the link between the current node and the other node
        :param node: izigraph.node.Node
        :return:izigraph.link.Link
        """
        if self.is_linked_to(node):
            return self.__links[node.label()]
        return None

    def unlink_from(self, node):
        """
        Removes the link between the current node and another node
        :param node: izigraph.node.Node
        """
        if self.is_linked_to(node):
            del self.__links[node.label()]
        else:
            raise Exception("Nodes are not linked")
